/* Define GPFS_64BIT_INODES to map the default interface definitions
   to 64-bit interfaces. Without this define, the 32-bit interface
   is the default.
*/
#define GPFS_64BIT_INODES
#include <gpfs.h>
#include <gpfs_fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

typedef long long Int64;

typedef struct 
{
  gpfsFcntlHeader_t hdr;
  gpfsGetDataBlkDiskIdx_t data;
} getDataBlk_t;

/* Utility function to support scale suffix (k|m|g) */
Int64 strToLongLong(char **argv, int argc, int i, const char* parameterName)
{
  Int64 value=0;
  int valLen;
  int intScale=1;
  char *str;
  char c;

  if (i > argc - 1)
  {
    printf("Missing %s\n", parameterName);
    exit(1);
  }

  str = argv[i];

  valLen = strlen(str);
  if (valLen > 0)
  {
    c = str[valLen-1];
    if ((c == 'k') || (c == 'K'))
      intScale = 1024;
    else if ((c == 'm') || (c == 'M'))
      intScale = 1024*1024;
    else if ((c == 'g') || (c == 'G'))
      intScale = 1024*1024*1024;

    value = atoll(str);
    value = value * intScale;
  }

  if (value < 0)
  {
    printf("Invalid value: %s must >= 0\n", parameterName);
    exit(1);
  } 
  return value;
}

int main(int argc, char **argv)
{
  getDataBlk_t gDB;
  int fd;
  int rc = 0;
  int i = 0;
  int j = 0;
  int tsoffset;
  OffsetLoc *offLoc;
  struct stat statBuf;
  char *fileName;
  Int64 startOffset = 0;
  /* 0 means metablock size */
  Int64 skipfactor = 0;
  Int64 length = -1;
  Int64 fileLength = -1;
  /* 0 means all replicas */
  int replicas = 0;
  Int64 lengthReturned = 0;
  Int64 offsetReturned = 0;
  int blockNum = 0;
  bool v2 = false;
  Int64 metaBlockSize = 0;
  int *diskNumP, *cacheNodeP;

  if (argc == 2)
  {
    fileName = argv[1];
    startOffset = 0;
    skipfactor = 0;
  } 
  else if (argc > 2) 
  {
    fileName = argv[1];
    for (i = 2; i < argc; i++) 
    {
      if (strcmp(argv[i], "-s") == 0)
      {
        // start offset
        startOffset = strToLongLong(argv, argc, ++i, "startOffset");
      } 
      else if (strcmp(argv[i], "-f") == 0)
      {
        // skip factor
        skipfactor = strToLongLong(argv, argc, ++i, "skipFactor");
        if (skipfactor < 0 || (skipfactor > 0 && skipfactor < 512))
        {
          printf("Invalid value: skipFactor should be 0 or >= 512 \n");
          return 1;
        }
      }
      else if (strcmp(argv[i], "-l") == 0)
      {
        // length
        length = strToLongLong(argv, argc, ++i, "length");
      } 
      else if (strcmp(argv[i], "-r") == 0) 
      {
	i++;
        // replicas
        if (i > argc - 1)
        {
          printf("missing -r parameter\n");
          return 1;
        }
        replicas = atoi(argv[i]);
        if (replicas < 0)
        {
          printf("Invalid value: replicas must >= 0\n");
          return 1;
        }
      }
      else if(strcmp(argv[i], "-v2") == 0)
      {
        // use GPFS_FCNTL_GET_DATABLKLOC
        v2 = true;
      }
    }
  }
  else
  {
    printf("usage: tsGetDataBlk <filename> [-v2] [-s <startoffset>] [-f <skipfactor>] [-l <length>] [-r <replicas>]\n");
    printf("       k|K|m|M|g|G suffix is acceptable\n");
    return 1;
  }

  fd = open(argv[1], O_RDONLY);
  if (fd < 0)
  {
    printf("Cannot open file %s\n", argv[1]);
    return 1;
  }

  if (fstat(fd, &statBuf) != 0) 
  {
    printf("Failed to stat %s\n", argv[1]);
    return 1;
  }
  fileLength = statBuf.st_size;

  if (length == -1)  
  {
    length = fileLength;
  }

  printf("File length: %lld, Block Size: %lld\n", fileLength, statBuf.st_blksize);
  if (skipfactor == 0)
    printf("Parameters: startoffset:%lld, skipfactor: META_BLOCK, length: %lld, replicas %d\n", 
            startOffset, length, replicas);
  else
    printf("Parameters: startoffset: %lld, skipfactor: %lld, length: %lld, replicas: %d\n", 
            startOffset, skipfactor, length, replicas);
  printf("\n");
 
  while (length > 0 && startOffset < fileLength)
  {
    gDB.hdr.totalLength = sizeof(gDB);
    gDB.hdr.fcntlVersion = GPFS_FCNTL_CURRENT_VERSION;
    gDB.hdr.fcntlReserved = 0;
    gDB.data.structLen = sizeof(gDB.data);
    if (!v2)
      gDB.data.structType = GPFS_FCNTL_GET_DATABLKDISKIDX;
    else
      gDB.data.structType = GPFS_FCNTL_GET_DATABLKLOC;

    gDB.data.filemapIn.startOffset = startOffset;
    gDB.data.filemapIn.skipfactor = skipfactor;
    gDB.data.filemapIn.length = length;
    gDB.data.filemapIn.mreplicas = replicas;


    rc = gpfs_fcntl(fd, &gDB); 
    if (rc < 0)
    {
      printf("gpfs_fcntl error:%d\n", rc);
      return 1;
    }
    
    lengthReturned = 0;

    if (!v2)
      metaBlockSize = gDB.data.filemapOut.blockSize;
    else
      metaBlockSize = ((Int64)gDB.data.filemapOut.blockSizeHigh << 32) + gDB.data.filemapOut.blockSize;

    printf("numReplicasReturned: %d, numBlksReturned: %d, META_BLOCK size: %lld\n",
            gDB.data.filemapOut.numReplicasReturned, 
            gDB.data.filemapOut.numBlksReturned, 
            metaBlockSize);
    if (gDB.data.filemapOut.numBlksReturned == 0)
      break;

    if (skipfactor == 0)
      skipfactor = metaBlockSize;

    offsetReturned = startOffset;
    for (i = 0; i < gDB.data.filemapOut.numBlksReturned; i++)
    {
      if (!v2)
      {
        tsoffset = sizeof(Int64) + (gDB.data.filemapOut.numReplicasReturned * sizeof(int));
        offLoc = (OffsetLoc*)(gDB.data.filemapOut.buffer + (i * tsoffset));
        printf("Block %d (offset %lld) is located at disks:", blockNum, offLoc->offset);

        for (j = 0; j < gDB.data.filemapOut.numReplicasReturned; j++)
        {
          printf("%3d ", offLoc->diskNum[j]);
        }
        printf("\n");

        lengthReturned += skipfactor;
        offsetReturned = offLoc->offset + skipfactor;
        blockNum++;
      }
      else 
      {
        printf("Block %d (offset %lld) is located at disks:", blockNum, offsetReturned);

        tsoffset = gDB.data.filemapOut.numReplicasReturned * sizeof(int);
        diskNumP = (int*) (gDB.data.filemapOut.buffer + (i * tsoffset *2));

        for (j = 0; j < gDB.data.filemapOut.numReplicasReturned; j++)
        {
          printf("%3d ", *diskNumP++);
        }
 
        printf("  cached at nodes: ");
        cacheNodeP = (int*) (gDB.data.filemapOut.buffer + (i * tsoffset *2 + tsoffset));

        for (j = 0; j < gDB.data.filemapOut.numReplicasReturned; j++)
        {
          printf("%3d ", *cacheNodeP++);
        } 
        printf("\n");
        lengthReturned += skipfactor;
        offsetReturned += skipfactor;
        blockNum++;
      }
    }

    startOffset = offsetReturned;
    length = length - lengthReturned;
  }

  close(fd);
}
